# -*- coding: utf-8 -*-
# Generated by Django 1.11.11 on 2018-04-10 19:52
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('djaken', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Image',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('image_data', models.TextField(blank=True, editable=False)),
                ('encrypted_image_data', models.TextField(blank=True, editable=False)),
                ('encryption_iv', models.CharField(blank=True, editable=False, max_length=64)),
            ],
        ),
        migrations.AlterField(
            model_name='note',
            name='content',
            field=models.TextField(blank=True),
        ),
        migrations.AlterField(
            model_name='note',
            name='content_encrypted',
            field=models.TextField(blank=True, editable=False),
        ),
        migrations.AlterField(
            model_name='note',
            name='content_html',
            field=models.TextField(blank=True, editable=False),
        ),
        migrations.AlterField(
            model_name='note',
            name='encryption_iv',
            field=models.CharField(blank=True, editable=False, max_length=64),
        ),
        migrations.AddField(
            model_name='image',
            name='note',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='djaken.Note'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.11 on 2018-04-03 22:35
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Note',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('title', models.CharField(max_length=255)),
                ('content', models.TextField()),
                ('content_html', models.TextField(editable=False)),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('relevant', models.BooleanField(default=True)),
                ('is_encrypted', models.BooleanField(default=False, editable=False)),
                ('content_encrypted', models.TextField(editable=False)),
                ('encryption_iv', models.CharField(editable=False, max_length=64)),
                ('author', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'ordering': ['-relevant', '-modified'],
            },
        ),
    ]



from django.conf.urls import url

from . import views

app_name = 'djaken'

urlpatterns = [
    url(r'^$', views.login, name='login'),
    url(r'^logout/$', views.logout, name='logout'),
    url(r'^all_notes/$', views.all_notes, name='all_notes'),
    url(r'^all_notes/toggle_relevant/$', views.all_notes, kwargs={'toggle_relevant':True}, name='all_notes_toggle_relevant'),
    url(r'^all_notes/sort_by_relevant/$', views.all_notes, kwargs={'sort_by':'relevant'}, name='all_notes_sort_by_relevant'),
    url(r'^all_notes/sort_by_modified/$', views.all_notes, kwargs={'sort_by':'modified'}, name='all_notes_sort_by_modified'),
    url(r'^all_notes/sort_by_created/$', views.all_notes, kwargs={'sort_by':'created'}, name='all_notes_sort_by_created'),
    url(r'^new_note/$', views.new_note, name='new_note'),
    url(r'^view_note/(?P<pk>[0-9]+)', views.view_note, name='view_note'),
    url(r'^view_note/toggle_relevant/(?P<pk>[0-9]+)/$', views.view_note, kwargs={'toggle_relevant':True}, name='view_note_toggle_relevant'),
    url(r'^edit_note/(?P<pk>[0-9]+)/$', views.edit_note, name='edit_note'),
    url(r'^unlock_note/(?P<pk>[0-9]+)/$', views.unlock_note, name='unlock_note'),
    url(r'^save_note/(?P<pk>[0-9]+)/$', views.save_note, name='save_note'),
]



from django.utils.translation import ugettext as _, ugettext_lazy
from django.contrib.auth import logout as auth_logout, REDIRECT_FIELD_NAME
from django.core.urlresolvers import reverse
from django.shortcuts import render, get_object_or_404, redirect
from django.contrib import messages
from django.views.decorators.cache import never_cache
from django.conf import settings
from django.db.models import Q
import ast
from djaken.models import Note
from djaken.models import Image

"""
Message examples
        messages.info(request, 'Hello World')
        messages.warning(request, 'Oh my gosh!')
        messages.error(request, 'Oh no!')
"""

BRANDING_TITLE = getattr(settings, "DJAKEN_BRANDING_TITLE", ugettext_lazy('Djaken Notes'))

def get_user_is_staff(request):
    return request.user.is_staff

def get_user_is_logged_in(request):
    return request.user.is_active

def login(request, extra_context=None):

    if request.method == 'GET' and get_user_is_staff(request):
        # Already logged-in, redirect to notes listing page
        return redirect('djaken:all_notes')

    from django.contrib.auth.views import login
 
    # Since this module gets imported in the application's root package,
    # it cannot import models from other applications at the module level,
    # and django.contrib.admin.forms eventually imports User.

    from django.contrib.admin.forms import AdminAuthenticationForm

    context = {
        'site_title': ugettext_lazy('Djaken'),
        'site_header': BRANDING_TITLE, 
        'page_title': ugettext_lazy('Log in'),
        'user_is_staff': get_user_is_staff(request),
        'user_is_logged_in': get_user_is_logged_in(request),
        'login_error_text': ugettext_lazy('The username and password is incorrect, please try again.'),
        'app_path': request.get_full_path(), 
    }

    if (REDIRECT_FIELD_NAME not in request.GET and REDIRECT_FIELD_NAME not in request.POST):
        context[REDIRECT_FIELD_NAME] = reverse('djaken:all_notes', current_app='djaken')

    context.update(extra_context or {})

    defaults = {
        'extra_context': context,
        'authentication_form': AdminAuthenticationForm,
        'template_name': 'djaken/login.html',
    }

    request.current_app = 'djaken'
    return login(request, **defaults)

def logout(request, extra_context=None):

    context = {}
    context.update(extra_context or {})

    defaults = {
        'extra_context': context,
    }
 
    request.current_app = 'djaken'

    auth_logout(request)
    return redirect('djaken:login') 

@never_cache
def all_notes(request, **kwargs):

    options = dict(kwargs)  
    ##print("views.py:all_notes:- options = ", options)

    next_url = None
    render_html_path = 'djaken/all_notes.html'

    if get_user_is_logged_in(request):
        if request.method == 'GET':

            # Set some defaults
            sort_order = ['-relevant','-modified','-created']
            only_relevant = True
            search_text = ""
 
            query_string_flat = request.META['QUERY_STRING'].replace('%20', ' ').replace('%27',"'").replace('%5B','[').replace('%5D',']')
 
            if query_string_flat != "":
                query_strings = query_string_flat.split('&', query_string_flat.count('&'))

                for query_string in query_strings:
                    if 'sort_order=' in query_string:
                        try:
                            sort_order_string = query_string.split('=',1)[1]
                            sort_order = ast.literal_eval(sort_order_string)
                        except:
                            pass
                    if 'only_relevant=' in query_string:
                        try:
                            only_relevant_string = query_string.split('=',1)[1]
                            only_relevant = ast.literal_eval(only_relevant_string)
                        except:
                            pass
                    if 'search=' in query_string:
                        try:
                            search_string = query_string.split('=',1)[1]
                            search_text = str(search_string)
                        except:
                            pass

            if 'toggle_relevant' in options.keys():
                if only_relevant == True:
                    only_relevant = False
                else:
                    only_relevant = True

            if 'sort_by' in options.keys():
                sort_by = options['sort_by']
                if sort_by in sort_order:
                    sort_order.remove(sort_by)
                    sort_order.insert(0,"-" + sort_by)
                elif "-" + sort_by in sort_order:
                    sort_order.remove("-" + sort_by)
                    sort_order.insert(0,sort_by)

            if only_relevant == True:
                if search_text != "":
                    notes = Note.objects.filter( Q(relevant=True) & Q(author=request.user) & (Q(title__icontains=search_text) | Q(content__icontains=search_text)) ).order_by(*sort_order)
                else:
                    notes = Note.objects.filter(relevant=True, author=request.user).order_by(*sort_order)
            else:
                if search_text != "":
                    notes = Note.objects.filter( Q(author=request.user) & (Q(title__icontains=search_text) | Q(content__icontains=search_text)) ).order_by(*sort_order)
                else:
                    notes = Note.objects.filter(author=request.user).order_by(*sort_order)

            context = {
                'site_title': ugettext_lazy('Djaken'),
                'site_header': BRANDING_TITLE,
                'page_title': ugettext_lazy('All Notes'),
                'content_title': ugettext_lazy('All Notes'),
                'user_is_staff': get_user_is_staff(request),
                'user_is_logged_in': get_user_is_logged_in(request),
                'app_path': request.get_full_path(),
                'only_relevant': only_relevant,
                'sort_order': sort_order,
                'notes': notes,
                'search_text': search_text,
            }

            return render(request, render_html_path, context)

    else:
        return redirect('djaken:login')

@never_cache
def view_note(request, pk, **kwargs):

    options = dict(kwargs)  
    ##print("views.py:view_note:- options = ", options)

    next_url = None
    render_html_path = None
 
    if get_user_is_logged_in(request):
        if request.method == 'GET':

            note = get_object_or_404(Note, pk=pk, author=request.user)
            if 'toggle_relevant' in options.keys():
                if note.relevant == True:
                    note.relevant = False
                else:
                    note.relevant = True
                note.save(True)
            if note.is_encrypted:
                next_url = 'djaken/view_note.html'
                render_html_path = 'djaken/unlock_note.html'
            else:
                render_html_path = 'djaken/view_note.html'
            context = {
                'site_title': ugettext_lazy('Djaken'),
                'site_header': BRANDING_TITLE,
                'page_title': ugettext_lazy('View Note'),
                'content_title': note.title,
                'user_is_staff': get_user_is_staff(request),
                'user_is_logged_in': get_user_is_logged_in(request),
                'app_path': request.get_full_path(),
                'next_url': next_url,
                'note': note,
                'note_content_html': note.content_html,
            }
            return render(request, render_html_path, context)
    else:
        return redirect('djaken:login')

@never_cache
def edit_note(request, pk, **kwargs):

    options = dict(kwargs)
    ##print("views.py:edit_note:- options = ", options)

    next_url = None
    render_html_path = None

    if get_user_is_logged_in(request):
        if request.method == 'GET':

            note = get_object_or_404(Note, pk=pk, author=request.user)
            if 'toggle_relevant' in options.keys():
                if note.relevant == True:
                    note.relevant = False
                else:
                    note.relevant = True
                note.save(True)
            if note.is_encrypted:
                next_url = 'djaken/edit_note.html'
                render_html_path = 'djaken/unlock_note.html'
            else:
                render_html_path = 'djaken/edit_note.html'
            context = {
                'site_title': ugettext_lazy('Djaken'),
                'site_header': BRANDING_TITLE,
                'page_title': ugettext_lazy('Edit Note'),
                'content_title': note.title,
                'user_is_staff': get_user_is_staff(request),
                'user_is_logged_in': get_user_is_logged_in(request),
                'app_path': request.get_full_path(),
                'next_url': next_url,
                'note': note,
                'note_content': note.content,
            }
            return render(request, render_html_path, context)
    else:
        return redirect('djaken:login')

@never_cache
def unlock_note(request, pk, **kwargs):

    options = dict(kwargs)
    ##print("views.py:unlock_note:- options = ", options)

    next_url = None
    render_html_path = None

    if get_user_is_logged_in(request):

        if request.method == 'GET':

            return redirect('/view_note/' + pk + '/')

        if request.method == 'POST':

            note = get_object_or_404(Note, pk=pk, author=request.user)
            render_html_path = request.POST['next_url']
            encryption_key = request.POST['encrypt_pass']
            success, unencrypted_content, unencrypted_content_html = note.get_unencrypted_content(encryption_key)

            if success:
                if render_html_path == "djaken/edit_note.html":
                    page_title = ugettext_lazy('Edit Note')
                if render_html_path == "djaken/view_note.html":
                    page_title = ugettext_lazy('View Note')
            else:
                messages.error(request, ugettext_lazy("ERROR: Could not decrypt the note! Perhaps the supplied encryption password was wrong."))
                if render_html_path == "djaken/edit_note.html":
                    return redirect('djaken:edit_note', pk=pk)
                if render_html_path == "djaken/view_note.html":
                    return redirect('djaken:view_note', pk=pk)
 
            context = {
                'site_title': ugettext_lazy('Djaken'),
                'site_header': BRANDING_TITLE,
                'page_title': page_title,
                'content_title': note.title,
                'user_is_staff': get_user_is_staff(request),
                'user_is_logged_in': get_user_is_logged_in(request),
                'app_path': request.get_full_path(),
                'next_url': next_url,
                'note': note,
                'note_content': unencrypted_content,
                'note_content_html': unencrypted_content_html,
            }
            return render(request, render_html_path, context)
    else:
        return redirect('djaken:login')

@never_cache
def save_note(request, pk, **kwargs):

    options = dict(kwargs)
    ##print("views.py:save_note:- options = ", options)

    next_url = None

    if get_user_is_logged_in(request):

        if request.method == 'GET':

            return redirect('/view_note/' + pk + '/')

        if request.method == 'POST':

            note = get_object_or_404(Note, pk=pk, author=request.user)
            new_title = request.POST['note_title']
            new_content = request.POST['note_content']
            image_data = request.POST['preview_image_data']
            is_encrypted =  bool(request.POST.get('is_encrypted', False))
            encryption_key = request.POST['encrypt_pass']

            note.title = new_title
            note.is_encrypted = is_encrypted
            note.encryption_key = encryption_key

            if request.POST['attach_during_save'] == "True" and image_data is not "":
                image = note.image_set.create(image_data=image_data)
                note.image_data_dict[str(image.id)] = image_data
                new_content += "\r\n\r\n.. image:: [[[" + str(image.id) + "]]]\r\n    :alt: Image #" + str(image.id) + "\r\n\r\n"

            note.content = new_content
            note.save()

            messages.info(request, ugettext_lazy("Note saved successfully!"))
            if is_encrypted:
                messages.warning(request, ugettext_lazy("Don't forget the encryption password!"))
            if request.POST['continue_after_save'] == "True":
                return redirect('djaken:edit_note', pk=pk)
            else:
                return redirect('djaken:view_note', pk=pk)

    else:
        return redirect('djaken:login')

@never_cache
def new_note(request, **kwargs):

    options = dict(kwargs)
    ##print("views.py:new_note:- options = ", options)

    next_url = None
    render_html_path = 'djaken/edit_note.html'

    if get_user_is_logged_in(request):
        if request.method == 'GET':

            note = Note(title=ugettext_lazy('New note...'), author=request.user)
            note.save()
            messages.info(request, ugettext_lazy("New note created! Give it a new title."))

            context = {
                'site_title': ugettext_lazy('Djaken'),
                'site_header': BRANDING_TITLE,
                'page_title': ugettext_lazy('New Note'),
                'content_title': ugettext_lazy('New note...'),
                'user_is_staff': get_user_is_staff(request),
                'user_is_logged_in': get_user_is_logged_in(request),
                'app_path': request.get_full_path(),
                'note': note,
            }

            return render(request, render_html_path, context)
    else:
        return redirect('djaken:login')




from django.db import models
from django.contrib.auth.models import User
from django.conf import settings
from docutils.core import publish_parts
from docutils.utils import SystemMessage
from Crypto.Cipher import AES
from Crypto import Random
import base64
import hashlib

class Note(models.Model):

    author = models.ForeignKey('auth.User', on_delete=models.PROTECT)
    title = models.CharField(max_length=255)
    content = models.TextField(blank=True)
    content_html = models.TextField(blank=True, editable=False)                    # don't want to see this in Admin
    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)
    relevant = models.BooleanField(default=True)
    is_encrypted = models.BooleanField(default=False, editable=False)              # don't want to see this in Admin
    content_encrypted = models.TextField(blank=True, editable=False)               # don't want to see this in Admin
    encryption_iv = models.CharField(blank=True, max_length=64, editable=False)    # don't want to see this in Admin
    encryption_key = ""                                                            # Only used momentarily to pass the key to the save function
    image_data_dict = {}

    def __init__(self, *args, **kwargs):

        super(Note, self).__init__(*args, **kwargs)
        # Ensure non-DB-backed variables are initialized/reset.
        encryption_key = ""
        self.image_data_dict = {}

    def get_html_from_reST(self, reST_text):
 
        reST_content = reST_text

        try:
            html = publish_parts(reST_content, writer_name='html')['html_body']
        except SystemMessage as e:
            message = str(e.args[0]).replace('<',"&lt").replace('>',"&gt")
            html = '<div class=publish_error><strong><i>Sorry, the publishing of this note from reStructuredText ' \
                   'to HTML failed with the following message:</i></strong><br><br><pre>' \
                   + message + '</pre></div>'
        return html


    def save(self, saveRelevancyOnly=False, *args, **kwargs):

        if not saveRelevancyOnly:
            if self.is_encrypted:
                encryption_key = self.encryption_key
                self.encryption_iv, self.content_encrypted = self.encrypt(self.content, encryption_key)
                self.content = "*This note is encrypted.*"
                self.content_html = self.get_html_from_reST(self.content)
                images = self.image_set.all()
                for image in images:
                    if image.encryption_iv is not "":
                        image.image_data = self.image_data_dict[str(image.id)]
                    image.save()
            else:
                for image_id in self.image_data_dict:
                    image = self.image_set.get(pk=image_id)
                    image.image_data = self.image_data_dict[str(image_id)]
                    image.save()
                expanded_content = self.expand_image_data(self.content)
                self.content_html = self.get_html_from_reST(expanded_content)

            # Before saving, destroy temporary sensitive data
            self.encryption_key = ""
            self.image_data_dict = {}
        super(Note, self).save(*args, **kwargs)

    def expand_image_data(self, collapsed_content):

        markupText = collapsed_content
        tagVertices = []
        tagLHSpartA = ".. image:: "
        tagLHSpartB = "[[["
        tagLHS = tagLHSpartA + tagLHSpartB
        tagRHS = "]]]"
        len_tagLHSpartA = len(tagLHSpartA)
        len_tagLHSpartB = len(tagLHSpartB)
        len_tagLHS = len(tagLHS)
        len_tagRHS = len(tagRHS)

        # Build a list of vertices (begin and end index) of image tags (".. image:: [[[n]]]" where n is image_id) 
        for cIndex in range(len(markupText)):
            if ( (cIndex < (len(markupText) - (len_tagLHS + len_tagRHS))) and (markupText[cIndex : cIndex + len_tagLHS] == tagLHS) ):
                for cIndexTag in range(len(markupText) - cIndex - len_tagLHS):
                    cIndexOffset = cIndexTag + cIndex + len_tagLHS
                    if markupText[cIndexOffset : cIndexOffset + len_tagRHS] == "]]]":
                        tagVertices.append((cIndex + len_tagLHSpartA, cIndexOffset + len_tagRHS))
                        break

        # Replace image tag id placeholders with image_data from database
        vertexOffset = 0
        for vertex in tagVertices:
            vertexBegin = vertex[0] + vertexOffset
            vertexEnd = vertex[1] + vertexOffset
            image_id_string = markupText[ vertexBegin : vertexEnd ]
            image_id = int(image_id_string[len_tagLHSpartB:len(image_id_string)-len_tagRHS])

            if self.is_encrypted:
                image_data = self.image_data_dict[str(image_id)]
            else:
                try:
                    image = self.image_set.get(pk=image_id)
                except (KeyError, Image.DoesNotExist):
                    image_data = broken_image_icon
                else:
                    image_data = image.image_data
            markupText = markupText[0:vertexBegin] + image_data + markupText[vertexEnd:len(markupText)]
            vertexOffset += len(image_data) - len(image_id_string)

        expanded_content = markupText
        return expanded_content

    def get_unencrypted_content(self, encryption_key):

        collapsed_unencrypted_content = self.decrypt(self.content_encrypted, encryption_key, self.encryption_iv)
        if collapsed_unencrypted_content:
            images = self.image_set.all()
            for image in images:
                unencrypted_image_data = self.decrypt(image.encrypted_image_data, encryption_key, image.encryption_iv)
                if (unencrypted_image_data):
                    self.image_data_dict [str(image.id)] = unencrypted_image_data
                else:
                    self.image_data_dict [str(image.id)] = "Decrypt failed for image id = " + str(image.id)
            unencrypted_content = self.expand_image_data(collapsed_unencrypted_content)
            unencrypted_content_html = self.get_html_from_reST(unencrypted_content)
            return True, collapsed_unencrypted_content, unencrypted_content_html
        else:
            return False, "", ""


    def encrypt(self, raw_text, raw_key):

        key = hashlib.md5(raw_key.encode('utf-8')).digest()
        iv = Random.new().read(AES.block_size)
        iv_encoded = base64.b64encode(iv)
        cipher = AES.new( key, AES.MODE_CFB, iv) 
        encrypted_bytes = cipher.encrypt(raw_text.encode('utf-8'))
        encrypted_text_encoded = base64.b64encode(encrypted_bytes)

        return iv_encoded, encrypted_text_encoded


    def decrypt(self, encrypted_text_encoded, raw_key, iv_encoded):

        try:
            key = hashlib.md5(raw_key.encode('utf-8')).digest()
            iv = base64.b64decode(iv_encoded)
            encrypted_bytes = base64.b64decode(encrypted_text_encoded)
            cipher = AES.new( key, AES.MODE_CFB, iv) 
            decrypted_text_encoded = cipher.decrypt(encrypted_bytes)
            decrypted_text = decrypted_text_encoded.decode('utf-8')

            return decrypted_text
        except:
            return False

    class Meta:
        ordering = ['-relevant', '-modified']


class Image(models.Model):

    note = models.ForeignKey(Note, on_delete=models.CASCADE)
    image_data = models.TextField(blank=True, editable=False)                      # don't want to see this in Admin
    encrypted_image_data = models.TextField(blank=True, editable=False)            # don't want to see this in Admin
    encryption_iv = models.CharField(blank=True, max_length=64, editable=False)    # don't want to see this in Admin

    def save(self, *args, **kwargs):
        if self.note.is_encrypted:
            self.encryption_iv, self.encrypted_image_data = self.note.encrypt(self.image_data, self.note.encryption_key)
            self.image_data = "*This_image_is_encrypted*"
        else:
            self.encrypted_image_data = ""
            self.encryption_iv = ""
        super(Image, self).save(*args, **kwargs)



broken_image_icon='data:image/png;base64, '



from django.contrib import admin
from djaken.models import Note

class FilterUserAdmin(admin.ModelAdmin): 

    def get_queryset(self, request): 
        qs = super(FilterUserAdmin, self).get_queryset(request)
        if request.user.is_superuser:
            return qs
        else:
            return qs.filter(author = request.user)


    def save_model(self, request, obj, form, change):
        obj.author = request.user
        obj.save()


    def has_change_permission(self, request, obj=None):
        if not obj:
            return True    # So they can see the change list page
        if request.user.is_superuser or obj.author == request.user:
            return True
        else:
            return False


    has_delete_permission = has_change_permission

class NoteAdmin(FilterUserAdmin):
    exclude = ("author",)
    list_display = ['title', 'created', 'modified', 'relevant', 'author']
    list_filter = ['title', 'created', 'modified', 'relevant', ('author', admin.RelatedOnlyFieldListFilter)]
    search_fields = ['title','content']
    date_hierarchy = 'created'
    #save_on_top = True

admin.site.register(Note, NoteAdmin)
